import 'package:json_path/src/expression/expression.dart';
import 'package:json_path/src/expression/nodes.dart';
import 'package:json_path/src/fun/fun.dart';
import 'package:json_path/src/fun/fun_call.dart';
import 'package:json_path/src/grammar/fun_name.dart';
import 'package:maybe_just_nothing/maybe_just_nothing.dart';
import 'package:petitparser/petitparser.dart';

class FunFactory {
  FunFactory(Iterable<NamedFun> functions) {
    for (final f in functions) {
      _validateName(f.name);
      final isValid = f is Fun<Maybe> || f is Fun<bool> || f is Fun<Nodes>;
      if (isValid || (f is Fun1 && _isValid1(f))) {
        (_fun[f.name] ??= []).add(f);
      } else {
        throw ArgumentError('Invalid function type for $f : ${f.runtimeType}');
      }
    }
  }

  bool _isValid1<A1>(Fun1<dynamic, A1> f) {
    return A1 == Maybe || A1 == Nodes || A1 == bool;
  }

  static void _validateName(String name) {
    if (!funName.allMatches(name).contains(name)) {
      throw ArgumentError('Invalid function name $name');
    }
  }

  final _fun = <String, List<NamedFun>>{};

  /// Returns a function to use in comparable context.
  Expression<Maybe>? comparable(FunCall call) =>
      _mapper(call.args, _byName(call.name).whereType<Fun<Maybe>>());

  /// Returns a function to use in logical context.
  Expression<bool>? logical(FunCall call) =>
      _mapper(call.args, _byName(call.name).whereType<Fun<bool>>());

  List<NamedFun> _byName(String name) => _fun[name] ?? [];

  /// Returns a function to use in any context.
  Expression? any(FunCall call) =>
      _mapper(call.args, _byName(call.name).whereType<Fun>());

  Expression<R>? _mapper<R>(List<Expression> args, Iterable<Fun<R>> list) {
    for (final fun in list) {
      try {
        return fun.toExpression(args);
      } on Exception {
        continue;
      }
    }
    return null;
  }
}
